/*
 * Copyright(c) 1992 Bell Communications Research, Inc. (Bellcore)
 * Copyright(c) 1995-99 Andrew Lister
 * Copyright  1999, 2000, 2001, 2002, 2003 by the LessTif Developers.
 *
 *                        All rights reserved
 * Permission to use, copy, modify and distribute this material for
 * any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies, and that the name of Bellcore not be used in advertising
 * or publicity pertaining to this material without the specific,
 * prior written permission of an authorized representative of
 * Bellcore.
 *
 * BELLCORE MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES, EX-
 * PRESS OR IMPLIED, WITH RESPECT TO THE SOFTWARE, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR ANY PARTICULAR PURPOSE, AND THE WARRANTY AGAINST IN-
 * FRINGEMENT OF PATENTS OR OTHER INTELLECTUAL PROPERTY RIGHTS.  THE
 * SOFTWARE IS PROVIDED "AS IS", AND IN NO EVENT SHALL BELLCORE OR
 * ANY OF ITS AFFILIATES BE LIABLE FOR ANY DAMAGES, INCLUDING ANY
 * LOST PROFITS OR OTHER INCIDENTAL OR CONSEQUENTIAL DAMAGES RELAT-
 * ING TO THE SOFTWARE.
 *
 * $Id: ScrollMgr.c,v 1.51 2004/08/04 06:11:16 dannybackx Exp $
 */

#ifdef HAVE_CONFIG_H
#include <XbaeConfig.h>
#endif

#include <stdio.h>

#include <Xm/Xm.h>
#include <Xm/ScrollBar.h>

#include <Xbae/MatrixP.h>
#include <Xbae/Draw.h>
#include <Xbae/Shadow.h>
#include <Xbae/ScrollMgr.h>
#include <Xbae/Utils.h>

#include <XbaeDebug.h>


/*
 * ScrollMgr implementation.
 * When we scroll using XCopyArea, occluding windows will cause GraphicsExpose
 * events to be generated, if there are no occluding windows then NoExpose
 * events will be generated. The removal of occluding windows will cause Expose
 * events.  If a number of scrolls (XCopyAreas) occur in quick succession,
 * the events will contain obsolete x/y information since our internal
 * coordinates have been scrolled to a new location.  The ScrollMgr
 * keeps track of scrolls and offsets required to relocate the events to the
 * current coordinate system.
 *
 * The Matrix widget has two ScrollMgrs, one for the Matrix's window
 * and one for the Clip widget's window.
 *
 * Each widgets compress_exposures field should be XtExposeCompressSeries
 * or XtExposeNoCompress.
 *
 * The idea behind this code is based on the PanHandler posted by Chuck Ocheret
 * (chuck@fid.morgan.com)
 */


/*
 * Create and initialize a ScrollMgr
 */
SmScrollMgr xbaeSmCreateScrollMgr(void)
{
        SmScrollMgr scrollMgr = XtNew(SmScrollMgrRec);

        scrollMgr->offset_x = 0;
        scrollMgr->offset_y = 0;
        scrollMgr->scroll_count = 0;
        scrollMgr->scroll_queue = NULL;
        scrollMgr->scrolling = False;

        return scrollMgr;
}

/*
 * Destroy a ScrollMgr, including any queued scrolls
 */
void xbaeSmDestroyScrollMgr(SmScrollMgr scrollMgr)
{
        if (scrollMgr->scroll_queue) {
                SmScrollNode node = scrollMgr->scroll_queue->next;

                while (node != scrollMgr->scroll_queue) {
                        SmScrollNode d = node;

                        node = node->next;
                        XtFree((char *) d);
                }
                XtFree((char *) node);
        }

        XtFree((char *) scrollMgr);
}

/*
 * Record a new scroll request in the ScrollMgr
 */
static void xbaeSmAddScroll(SmScrollMgr scrollMgr, int delta_x, int delta_y)
{
        SmScrollNode node = XtNew(SmScrollNodeRec);

        node->x = delta_x;
        node->y = delta_y;

        scrollMgr->offset_x += delta_x;
        scrollMgr->offset_y += delta_y;
        scrollMgr->scroll_count++;

        /*
         * Insert the node at the end of the queue
         */
        if (!scrollMgr->scroll_queue) {
                scrollMgr->scroll_queue = node;
                node->next = node;
                node->prev = node;
        } else {
                SmScrollNode last = scrollMgr->scroll_queue->prev;

                last->next = node;
                node->next = scrollMgr->scroll_queue;
                node->prev = last;
                scrollMgr->scroll_queue->prev = node;
        }
}

/*
 * Remove a scroll from the ScrollMgr queue
 */
static void xbaeSmRemoveScroll(SmScrollMgr scrollMgr)
{
        if (scrollMgr->scroll_count) {
                SmScrollNode node = scrollMgr->scroll_queue;

                scrollMgr->offset_x -= node->x;
                scrollMgr->offset_y -= node->y;

                /*
                 * Remove node from head of queue
                 */
                if (node->next == node)
                        scrollMgr->scroll_queue = NULL;
                else {
                        scrollMgr->scroll_queue = node->next;
                        node->next->prev = node->prev;
                        node->prev->next = node->next;
                }
                XtFree((char *) node);

                scrollMgr->scroll_count--;
        }
}

/*
 * Handle an expose event
 */
Boolean xbaeSmScrollEvent(SmScrollMgr scrollMgr, XEvent * event, Rectangle * expose)
{
        switch (event->type) {

        case Expose:
                /*
                 * Normal Expose event, translate it into our scrolled
                 * coordinate system.
                 */
                DEBUGOUT(_XbaeDebug(__FILE__, NULL, "xbaeSmScrollEvent: Expose\n"));

                /*
                 * Translate the event into our scrolled coordinate system.
                 */
                SETRECT(*expose, event->xexpose.x, event->xexpose.y,
                        event->xexpose.x + event->xexpose.width - 1,
                        event->xexpose.y + event->xexpose.height - 1);

                return True;

        case GraphicsExpose:
                /*
                 * If we are not scrolling, then this must be the first
                 * GraphicsExpose event.  Remove the corresponding scroll from the
                 * queue, and if we have more GraphicsExposes to come, set scrolling
                 * to True.
                 */
                DEBUGOUT(_XbaeDebug(__FILE__, NULL, "xbaeSmScrollEvent: GraphicsExpose\n"));

                if (scrollMgr->scrolling == False) {
                        xbaeSmRemoveScroll(scrollMgr);
                        if (event->xgraphicsexpose.count != 0)
                                scrollMgr->scrolling = True;
                }

                /*
                 * This is the last GraphicsExpose so set scrolling to False.
                 */
                else if (event->xgraphicsexpose.count == 0)
                        scrollMgr->scrolling = False;

                /*
                 * Translate the event into our scrolled coordinate system.
                 */
                SETRECT(*expose, event->xgraphicsexpose.x, event->xgraphicsexpose.y,
                        event->xgraphicsexpose.x + event->xgraphicsexpose.width - 1,
                        event->xgraphicsexpose.y + event->xgraphicsexpose.height - 1);

                return True;

        case NoExpose:

                /*
                 * A NoExpose event means we won't be getting any GraphicsExpose
                 * events, so remove the scroll from the queue and set scrolling
                 * to False.
                 */
                xbaeSmRemoveScroll(scrollMgr);
                scrollMgr->scrolling = False;
                break;

        default:
                break;
        }

        return False;
}

static void
xbaeSmScrollRect(SmScrollMgr scrollMgr, Rectangle * expose, Rectangle * scrolled_expose, int clip,
                 Boolean scrolled)
{
        *scrolled_expose = *expose;

        if (scrolled) {
                if (clip & CLIP_VISIBLE_WIDTH) {
                        scrolled_expose->x1 += scrollMgr->offset_x;
                        scrolled_expose->x2 += scrollMgr->offset_x;
                }

                if (clip & CLIP_VISIBLE_HEIGHT) {
                        scrolled_expose->y1 += scrollMgr->offset_y;
                        scrolled_expose->y2 += scrollMgr->offset_y;
                }
        }
}

/*
 * Callback for vertical scrollbar
 * SGO: changed scrolling routine for vertical scrolling.
 * using same approach as in xbaeScrollHorizCB.
 */

/* ARGSUSED */
void xbaeScrollVertCB(Widget w, XtPointer client_data, XmScrollBarCallbackStruct * call_data)
{
        XbaeMatrixWidget mw = (XbaeMatrixWidget) XtParent(w);
        Rectangle fixed, nonfixed;
        int src_y, dest_y, copy_height;
        int y_clear, clear_height;
        int vert_sb_offset = VERT_SB_OFFSET(mw);
        int trailing_fixed_column_label_offset = TRAILING_FIXED_COLUMN_LABEL_OFFSET(mw);
        int trailing_fixed_column_width = TRAILING_FIXED_COLUMN_WIDTH(mw);
        int row_label_width = ROW_LABEL_WIDTH(mw);
        int fixed_row_label_offset = FIXED_ROW_LABEL_OFFSET(mw);
        int horiz_dead_space_width = HORIZ_DEAD_SPACE_WIDTH(mw);
        Boolean need_horiz_fill = NEED_HORIZ_FILL(mw);
        Boolean has_attached_trailing_columns = HAS_ATTACHED_TRAILING_COLUMNS(mw);
        Boolean need_horiz_dead_space_fill = NEED_HORIZ_DEAD_SPACE_FILL(mw);

        if (call_data->value == VERT_ORIGIN(mw))
                /* Didn't scroll */
                return;
        else if (call_data->value > VERT_ORIGIN(mw)) {
                /* scrolled down */
                dest_y = 0;
                src_y = call_data->value - VERT_ORIGIN(mw);
                copy_height = ClipChild(mw)->core.height - src_y;
                y_clear = copy_height;
                clear_height = ClipChild(mw)->core.height - copy_height;
        } else {
                /* scrolled up */
                dest_y = VERT_ORIGIN(mw) - call_data->value;
                src_y = 0;
                copy_height = ClipChild(mw)->core.height - dest_y;
                y_clear = 0;
                clear_height = ClipChild(mw)->core.height - copy_height;
        }

        /*
         * The textField needs to scroll along with the cells.
         */
        if (XtIsManaged(TextChild(mw)) && mw->matrix.current_row >= (int) mw->matrix.fixed_rows
            && mw->matrix.current_row < TRAILING_VERT_ORIGIN(mw)) {
                XtMoveWidget(TextChild(mw), TextChild(mw)->core.x,
                             TextChild(mw)->core.y + (VERT_ORIGIN(mw) - call_data->value));
        }

        /* Cell widgets that are in the vertical scrolling region
         * need to scroll along as well... Linas */
        if (mw->matrix.per_cell) {
                int roe, kol;

                for (kol = 0; kol < (mw->matrix.columns); kol++) {
                        for (roe = mw->matrix.fixed_rows; roe < TRAILING_VERT_ORIGIN(mw); roe++) {
                                Widget w = mw->matrix.per_cell[roe][kol].widget;
                                if (w) {
                                        if (XtIsManaged(w)) {
                                                int x = w->core.x;
                                                int y = w->core.y;
                                                y += VERT_ORIGIN(mw) - call_data->value;
                                                XtMoveWidget(w, x, y);
                                                xbaeMoveWindowFromUserWidgetXY(mw, w, roe, kol, x,
                                                                               y);
                                        }
                                }
                        }
                }
        }

        /*
         * Now we can adjust our vertical origin
         */
        VERT_ORIGIN(mw) = call_data->value;
        TOP_ROW(mw) = xbaeTopRow(mw) - mw->matrix.fixed_rows;

        if (!XtIsRealized((Widget) mw))
                return;

        if (copy_height <= 0) {
                /* nothing to copy, clear and redraw everything */
                y_clear = 0;
                clear_height = ClipChild(mw)->core.height;
        } else {
                /* There is something that can be copied */

                Boolean queue = False;
                int unattached_trailing_columns_offset = UNATTACHED_TRAILING_COLUMNS_OFFSET(mw);

                /*
                 * Copy the non-fixed cells in the clip widget
                 */
                XCopyArea(XtDisplay(mw), XtWindow(ClipChild(mw)), XtWindow(ClipChild(mw)),
                          mw->matrix.draw_gc, 0, src_y, VISIBLE_WIDTH(mw), copy_height, 0, dest_y);

                /*
                 * Queue this scroll with the clip ScrollMgr
                 */
                xbaeSmAddScroll(mw->matrix.clip_scroll_mgr, 0, dest_y - src_y);

                /*
                 * Copy cells in the fixed columns on LeftClip
                 */
                if (XtIsManaged(LeftClip(mw))) {
                        queue = True;
                        XCopyArea(XtDisplay(mw), XtWindow(LeftClip(mw)), XtWindow(LeftClip(mw)),
                                  mw->matrix.draw_gc, 0, src_y, FIXED_COLUMN_WIDTH(mw), copy_height,
                                  0, dest_y);
                }

                /*
                 * Copy cells in the trailing fixed columns on RightClip
                 */
                if (XtIsManaged(RightClip(mw))) {
                        queue = True;
                        /* Tobias: Not symetrical (horiz_sb_offset) */
                        XCopyArea(XtDisplay(mw), XtWindow(RightClip(mw)), XtWindow(RightClip(mw)),
                                  mw->matrix.draw_gc, 0, src_y, TRAILING_FIXED_COLUMN_WIDTH(mw),
                                  copy_height, 0, dest_y);
                }

                /*
                 * Translate coordinates to be relative to the matrix
                 */
                src_y += fixed_row_label_offset;
                dest_y += fixed_row_label_offset;

                /*
                 * Copy the row labels
                 */
                if (row_label_width) {
                        queue = True;
                        XCopyArea(XtDisplay(mw), XtWindow(mw), XtWindow(mw), mw->matrix.draw_gc,
                                  vert_sb_offset, src_y, row_label_width, copy_height,
                                  vert_sb_offset, dest_y);
                }

                /*
                 * Copy trailing filled portion of the rows if necessary
                 */
                if (need_horiz_fill && (!has_attached_trailing_columns)) {
                        queue = True;
                        XCopyArea(XtDisplay(mw), XtWindow(mw), XtWindow(mw), mw->matrix.draw_gc,
                                  trailing_fixed_column_label_offset + trailing_fixed_column_width,
                                  src_y, FILL_HORIZ_WIDTH(mw), copy_height,
                                  trailing_fixed_column_label_offset + trailing_fixed_column_width,
                                  dest_y);
                }
                if (need_horiz_dead_space_fill) {
                        queue = True;
                        XCopyArea(XtDisplay(mw), XtWindow(mw), XtWindow(mw), mw->matrix.draw_gc,
                                  unattached_trailing_columns_offset, src_y, horiz_dead_space_width,
                                  copy_height, unattached_trailing_columns_offset, dest_y);
                }

                /*
                 * Queue this scroll with the matrix ScrollMgr only if there were copies 
                 */
                if (queue) {
                        xbaeSmAddScroll(mw->matrix.matrix_scroll_mgr, 0, dest_y - src_y);
                }
        }

        /* Clear the areas that can't be copied */

        XClearArea(XtDisplay(mw), XtWindow(ClipChild(mw)), 0, y_clear, 0, clear_height, False);

        if (XtIsManaged(LeftClip(mw)))
                XClearArea(XtDisplay(mw), XtWindow(LeftClip(mw)), 0, y_clear, 0, clear_height,
                           False);

        if (XtIsManaged(RightClip(mw)))
                XClearArea(XtDisplay(mw), XtWindow(RightClip(mw)), 0, y_clear, 0, clear_height,
                           False);

        if (row_label_width)
                XClearArea(XtDisplay(mw), XtWindow(mw), vert_sb_offset,
                           fixed_row_label_offset + y_clear, row_label_width, clear_height, False);

        if (need_horiz_fill && (!has_attached_trailing_columns))
                XClearArea(XtDisplay(mw), XtWindow(mw),
                           trailing_fixed_column_label_offset + trailing_fixed_column_width,
                           fixed_row_label_offset + y_clear, FILL_HORIZ_WIDTH(mw), clear_height,
                           False);

        if (need_horiz_dead_space_fill)
                XClearArea(XtDisplay(mw), XtWindow(mw), UNATTACHED_TRAILING_COLUMNS_OFFSET(mw),
                           fixed_row_label_offset + y_clear, horiz_dead_space_width, clear_height,
                           False);

        SETRECT(nonfixed, 0, y_clear, ClipChild(mw)->core.width - 1, y_clear + clear_height - 1);

        SETRECT(fixed, vert_sb_offset, fixed_row_label_offset + y_clear,
                trailing_fixed_column_label_offset + trailing_fixed_column_width - 1,
                fixed_row_label_offset + y_clear + clear_height - 1);

        /*
         * Perform the actual redraw. The call to draw the grid shadows
         * must be done after any XClearAreas() to ensure we don't redraw
         * more than we have to; that is, we could put the call in the
         * redraw cells routine, but we would end up occasionally redrawing
         * more than once.
         */
        xbaeRedrawLabelsAndFixed(mw, &fixed);
        xbaeRedrawCells(mw, &nonfixed);
}

/*
 * Callback for horizontal scrollbar
 */

/* ARGSUSED */
void xbaeScrollHorizCB(Widget w, XtPointer client_data, XmScrollBarCallbackStruct * call_data)
{
        XbaeMatrixWidget mw = (XbaeMatrixWidget) XtParent(w);
        Rectangle fixed, nonfixed;
        int src_x, dest_x, copy_width;
        int x_clear, clear_width;
        int horiz_sb_offset = HORIZ_SB_OFFSET(mw);
        int trailing_fixed_row_label_offset = TRAILING_FIXED_ROW_LABEL_OFFSET(mw);
        int trailing_fixed_row_height = TRAILING_FIXED_ROW_HEIGHT(mw);
        int column_label_height = COLUMN_LABEL_HEIGHT(mw);
        int fixed_column_label_offset = FIXED_COLUMN_LABEL_OFFSET(mw);
        int vert_dead_space_height = VERT_DEAD_SPACE_HEIGHT(mw);
        Boolean need_vert_fill = NEED_VERT_FILL(mw);
        Boolean has_attached_trailing_rows = HAS_ATTACHED_TRAILING_ROWS(mw);
        Boolean need_vert_dead_space_fill = NEED_VERT_DEAD_SPACE_FILL(mw);

        if (call_data->value == HORIZ_ORIGIN(mw))
                /* Didn't scroll */
                return;
        else if (call_data->value > HORIZ_ORIGIN(mw)) {
                /* Scrolled right */
                dest_x = 0;
                src_x = call_data->value - HORIZ_ORIGIN(mw);
                copy_width = ClipChild(mw)->core.width - src_x;
                x_clear = copy_width;
                clear_width = ClipChild(mw)->core.width - copy_width;
        } else {
                /* Scrolled left */
                dest_x = HORIZ_ORIGIN(mw) - call_data->value;
                src_x = 0;
                copy_width = ClipChild(mw)->core.width - dest_x;
                x_clear = 0;
                clear_width = ClipChild(mw)->core.width - copy_width;
        }

        /*
         * The textField needs to scroll along with the cells.
         */
        if (XtIsManaged(TextChild(mw))
            && mw->matrix.current_column >= (int) mw->matrix.fixed_columns
            && mw->matrix.current_column < TRAILING_HORIZ_ORIGIN(mw)) {
                XtMoveWidget(TextChild(mw),
                             TextChild(mw)->core.x + (HORIZ_ORIGIN(mw) - call_data->value),
                             TextChild(mw)->core.y);
                DEBUGOUT(_XbaeDebug2
                         (__FILE__, (Widget) mw, w, "xbaeScrollVertCB: move child to %d %d\n",
                          TextChild(mw)->core.x, TextChild(mw)->core.y));
        }

        /* Cell widgets that are in the horizontal scrolling region
         * need to scroll along as well... Linas */
        if (mw->matrix.per_cell) {
                int roe, kol;

                for (roe = 0; roe < (mw->matrix.rows); roe++) {
                        for (kol = mw->matrix.fixed_columns; kol < TRAILING_HORIZ_ORIGIN(mw); kol++) {
                                Widget w = mw->matrix.per_cell[roe][kol].widget;
                                if (w) {
                                        if (XtIsManaged(w)) {
                                                int x = w->core.x;
                                                int y = w->core.y;
                                                x += HORIZ_ORIGIN(mw) - call_data->value;
                                                XtMoveWidget(w, x, y);

                                                DEBUGOUT(_XbaeDebug2
                                                         (__FILE__, (Widget) mw, w,
                                                          "xbaeScrollVertCB: move child to %d %d\n",
                                                          x, y));
                                                /* Need to call XMoveWindow after every call to
                                                 * XtMoveWidget, to make sure the widget appears in the
                                                 * correct place, and Intrinsics recognises the new
                                                 * coordinates correctly - Phil Eccles 10/04/2002 */
                                                xbaeMoveWindowFromUserWidgetXY(mw, w, roe, kol, x,
                                                                               y);
                                        }
                                }
                        }
                }
        }

        /*
         * Now we can adjust our horizontal origin
         */
        HORIZ_ORIGIN(mw) = call_data->value;
        mw->matrix.left_column = xbaeLeftColumn(mw) - mw->matrix.fixed_columns;

        if (!XtIsRealized((Widget) mw))
                return;

        if (copy_width <= 0) {
                /* nothing to copy, clear and redraw everything */
                x_clear = 0;
                clear_width = ClipChild(mw)->core.width;
        } else {
                /* There is something that can be copied */

                Boolean queue = False;
                int unattached_trailing_rows_offset = UNATTACHED_TRAILING_ROWS_OFFSET(mw);

                /*
                 * Copy the non-fixed cells in the clip widget
                 */
                XCopyArea(XtDisplay(mw), XtWindow(ClipChild(mw)), XtWindow(ClipChild(mw)),
                          mw->matrix.draw_gc, src_x, 0, copy_width, VISIBLE_HEIGHT(mw), dest_x, 0);

                /*
                 * Queue this scroll with the clip ScrollMgr
                 */
                xbaeSmAddScroll(mw->matrix.clip_scroll_mgr, dest_x - src_x, 0);

                /*
                 * Copy cells across in fixed rows on TopClip
                 */
                if (XtIsManaged(TopClip(mw))) {
                        queue = True;
                        XCopyArea(XtDisplay(mw), XtWindow(TopClip(mw)), XtWindow(TopClip(mw)),
                                  mw->matrix.draw_gc, src_x, 0, copy_width, FIXED_ROW_HEIGHT(mw),
                                  dest_x, 0);
                }

                /*
                 * Copy cells across in trailing fixed rows on BottomClip
                 */
                if (XtIsManaged(BottomClip(mw))) {
                        queue = True;
                        XCopyArea(XtDisplay(mw), XtWindow(BottomClip(mw)), XtWindow(BottomClip(mw)),
                                  mw->matrix.draw_gc, src_x, 0, copy_width,
                                  TRAILING_FIXED_ROW_HEIGHT(mw), dest_x, 0);
                }

                /*
                 * Translate coordinates for column labels on the Matrix.
                 */
                src_x += fixed_column_label_offset;
                dest_x += fixed_column_label_offset;

                /*
                 * Copy the column labels
                 */
                if (column_label_height) {
                        queue = True;
                        XCopyArea(XtDisplay(mw), XtWindow(mw), XtWindow(mw), mw->matrix.draw_gc,
                                  src_x, horiz_sb_offset, copy_width, column_label_height, dest_x,
                                  horiz_sb_offset);
                }

                /*
                 * Copy trailing filled portion of the columns if necessary
                 */
                if (need_vert_fill && (!has_attached_trailing_rows)) {
                        queue = True;
                        XCopyArea(XtDisplay(mw), XtWindow(mw), XtWindow(mw), mw->matrix.draw_gc,
                                  src_x,
                                  trailing_fixed_row_label_offset + trailing_fixed_row_height,
                                  copy_width, FILL_VERT_HEIGHT(mw), dest_x,
                                  trailing_fixed_row_label_offset + trailing_fixed_row_height);
                }

                if (need_vert_dead_space_fill) {
                        queue = True;
                        XCopyArea(XtDisplay(mw), XtWindow(mw), XtWindow(mw), mw->matrix.draw_gc,
                                  src_x, unattached_trailing_rows_offset, copy_width,
                                  vert_dead_space_height, dest_x, unattached_trailing_rows_offset);
                }

                /*
                 * Queue this scroll with the matrix ScrollMgr only if there were copies 
                 */
                if (queue) {
                        xbaeSmAddScroll(mw->matrix.matrix_scroll_mgr, dest_x - src_x, 0);
                }
        }

        /* Clear the areas that can't be copied */

        XClearArea(XtDisplay(mw), XtWindow(ClipChild(mw)), x_clear, 0, clear_width, 0, False);

        if (XtIsManaged(TopClip(mw)))
                XClearArea(XtDisplay(mw), XtWindow(TopClip(mw)), x_clear, 0, clear_width, 0, False);

        if (XtIsManaged(BottomClip(mw)))
                XClearArea(XtDisplay(mw), XtWindow(BottomClip(mw)), x_clear, 0, clear_width, 0,
                           False);

        if (column_label_height)
                XClearArea(XtDisplay(mw), XtWindow(mw), fixed_column_label_offset + x_clear,
                           horiz_sb_offset, clear_width, column_label_height, False);

        if (need_vert_fill && (!has_attached_trailing_rows))
                XClearArea(XtDisplay(mw), XtWindow(mw), fixed_column_label_offset + x_clear,
                           trailing_fixed_row_label_offset + trailing_fixed_row_height, clear_width,
                           FILL_HORIZ_WIDTH(mw), False);

        if (need_vert_dead_space_fill)
                XClearArea(XtDisplay(mw), XtWindow(mw), fixed_column_label_offset + x_clear,
                           UNATTACHED_TRAILING_ROWS_OFFSET(mw), clear_width, vert_dead_space_height,
                           False);

        SETRECT(nonfixed, x_clear, 0, x_clear + clear_width - 1, ClipChild(mw)->core.height - 1);

        SETRECT(fixed, fixed_column_label_offset + x_clear, horiz_sb_offset,
                fixed_column_label_offset + x_clear + clear_width - 1,
                trailing_fixed_row_label_offset + trailing_fixed_row_height - 1);

        /*
         * Perform the actual redraw. The call to draw the grid shadows
         * must be done after any XClearAreas() to ensure we don't redraw
         * more than we have to; that is, we could put the call in the
         * redraw cells routine, but we would end up occasionally redrawing
         * more than once.
         */
        xbaeRedrawLabelsAndFixed(mw, &fixed);
        xbaeRedrawCells(mw, &nonfixed);
}

#define SIMPLE_X_INTERSECT(r1, r2, res) \
    do { \
        (res).x1 = Max((r1).x1, (r2).x1); \
        (res).x2 = Min((r1).x2, (r2).x2); \
    } while(0)

#define SIMPLE_Y_INTERSECT(r1, r2, res) \
    do { \
        (res).y1 = Max((r1).y1, (r2).y1); \
        (res).y2 = Min((r1).y2, (r2).y2); \
    } while(0)

#define SIMPLE_INTERSECT(r1, r2, res) \
    do { \
        SIMPLE_X_INTERSECT(r1, r2, res); \
        SIMPLE_Y_INTERSECT(r1, r2, res); \
    } while(0)

/*
 * Redraw all cells in the clip widget damaged by the passed Rectangle.
 * The Rectangle must be within the bounds of the cells. These are the
 * non-fixed cells.
 */
static void RedrawCells(XbaeMatrixWidget mw, Rectangle * expose, Boolean scrolled)
{
        Rectangle rect;

        if ((mw->matrix.disable_redisplay) || (!mw->matrix.rows) || (!mw->matrix.columns))
                return;

        Rectangle matrix_expose, scrolled_expose;
        SETRECT(matrix_expose, expose->x1 + FIXED_COLUMN_LABEL_OFFSET(mw),
                expose->y1 + FIXED_ROW_LABEL_OFFSET(mw), expose->x2 + FIXED_COLUMN_LABEL_OFFSET(mw),
                expose->y2 + FIXED_ROW_LABEL_OFFSET(mw));

        DEBUGOUT(_XbaeDebug
                 (__FILE__, (Widget) mw,
                  "xbaeRedrawCells() scroll_count=%d (offset_x,offset_y,count)=(%d,%d) with expose (x,y)=((%d,%d)-(%d,%d)), matrix_expose=((%d,%d)-(%d,%d))\n",
                  mw->matrix.clip_scroll_mgr->scroll_count, mw->matrix.clip_scroll_mgr->offset_x,
                  mw->matrix.clip_scroll_mgr->offset_y, expose->x1, expose->y1, expose->x2,
                  expose->y2, matrix_expose.x1, matrix_expose.y1, matrix_expose.x2,
                  matrix_expose.y2));

        xbaeSmScrollRect(mw->matrix.clip_scroll_mgr, &matrix_expose, &scrolled_expose,
                         CLIP_VISIBLE_WIDTH | CLIP_VISIBLE_HEIGHT, scrolled);

        if (xbaeSetClipRect(mw, &rect, CLIP_VISIBLE_WIDTH | CLIP_VISIBLE_HEIGHT)
            && OVERLAP(scrolled_expose, rect)) {

                Rectangle intersect;
                int row, start_row, end_row;
                int column, start_column, end_column;

                SIMPLE_INTERSECT(scrolled_expose, rect, intersect);

                DEBUGOUT(_XbaeDebug
                         (__FILE__, (Widget) mw,
                          "redrawing cells (x,y)=((%d,%d)-(%d,%d)) overlap (x,y)=((%d,%d)-(%d,%d))\n",
                          rect.x1, rect.y1, rect.x2, rect.y2, intersect.x1, intersect.y1,
                          intersect.x2, intersect.y2));

                xbaeMatrixYtoRow(mw, &intersect.y1, &start_row);
                xbaeMatrixYtoRow(mw, &intersect.y2, &end_row);

                xbaeMatrixXtoColumn(mw, &intersect.x1, &start_column);
                xbaeMatrixXtoColumn(mw, &intersect.x2, &end_column);

                DEBUGOUT(_XbaeDebug
                         (__FILE__, (Widget) mw, "redrawing cells (row,column)=((%d,%d)-(%d,%d))\n",
                          start_row, start_column, end_row, end_column));

                for (row = start_row; row <= end_row; row++) {
                        for (column = start_column; column <= end_column; column++) {
                                xbaeDrawCell(mw, row, column);
                        }
                }
        }
}
void xbaeRedrawCells(XbaeMatrixWidget mw, Rectangle * expose)
{
        RedrawCells(mw, expose, False);
}

void xbaeRedrawScrolledCells(XbaeMatrixWidget mw, Rectangle * expose)
{
        RedrawCells(mw, expose, True);
}

static void xbaeRedrawRowLabels(XbaeMatrixWidget mw, Rectangle * expose, Boolean scrolled)
{
        Rectangle scrolled_expose, rect, intersect;
        int region;
        int row, start_row, end_row;

        int row_label_regions[] = {
                CLIP_ROW_LABELS | CLIP_FIXED_ROWS,
                CLIP_ROW_LABELS | CLIP_VISIBLE_HEIGHT,
                CLIP_ROW_LABELS | CLIP_TRAILING_FIXED_ROWS
        };

        for (region = 0; region < sizeof row_label_regions / sizeof *row_label_regions; region++) {

                xbaeSmScrollRect(mw->matrix.matrix_scroll_mgr, expose, &scrolled_expose,
                                 row_label_regions[region], scrolled);

                if (xbaeSetClipRect(mw, &rect, row_label_regions[region])
                    && OVERLAP(scrolled_expose, rect)) {

                        SIMPLE_Y_INTERSECT(scrolled_expose, rect, intersect);

                        DEBUGOUT(_XbaeDebug
                                 (__FILE__, (Widget) mw,
                                  "redrawing row label region %d (x,y)=((%d,%d)-(%d,%d)) overlap (x,y)=((%d,%d)-(%d,%d))\n",
                                  region, rect.x1, rect.y1, rect.x2, rect.y2, intersect.x1,
                                  intersect.y1, intersect.x2, intersect.y2));

                        xbaeMatrixYtoRow(mw, &intersect.y1, &start_row);
                        xbaeMatrixYtoRow(mw, &intersect.y2, &end_row);

                        DEBUGOUT(_XbaeDebug
                                 (__FILE__, (Widget) mw,
                                  "redrawing row label region %d rows=(%d-%d)\n", region, start_row,
                                  end_row));

                        for (row = start_row; row <= end_row; row++) {
                                xbaeDrawRowLabel(mw, row, False);
                        }
                }
        }
}

static void xbaeRedrawColumnLabels(XbaeMatrixWidget mw, Rectangle * expose, Boolean scrolled)
{
        Rectangle scrolled_expose, rect, intersect;
        int region;
        int column, start_column, end_column;

        int column_label_regions[] = {
                CLIP_COLUMN_LABELS | CLIP_FIXED_COLUMNS,
                CLIP_COLUMN_LABELS | CLIP_VISIBLE_WIDTH,
                CLIP_COLUMN_LABELS | CLIP_TRAILING_FIXED_COLUMNS
        };

        for (region = 0; region < sizeof column_label_regions / sizeof *column_label_regions;
             region++) {

                xbaeSmScrollRect(mw->matrix.matrix_scroll_mgr, expose, &scrolled_expose,
                                 column_label_regions[region], scrolled);

                if (xbaeSetClipRect(mw, &rect, column_label_regions[region])
                    && OVERLAP(scrolled_expose, rect)) {

                        SIMPLE_X_INTERSECT(scrolled_expose, rect, intersect);

                        DEBUGOUT(_XbaeDebug
                                 (__FILE__, (Widget) mw,
                                  "redrawing column label region %d (x,y)=((%d,%d)-(%d,%d)) overlap (x,y)=((%d,%d)-(%d,%d))\n",
                                  region, rect.x1, rect.y1, rect.x2, rect.y2, intersect.x1,
                                  intersect.y1, intersect.x2, intersect.y2));

                        xbaeMatrixXtoColumn(mw, &intersect.x1, &start_column);
                        xbaeMatrixXtoColumn(mw, &intersect.x2, &end_column);

                        DEBUGOUT(_XbaeDebug
                                 (__FILE__, (Widget) mw,
                                  "redrawing column label region %d columns=(%d-%d)\n", region,
                                  start_column, end_column));

                        for (column = start_column; column <= end_column; column++) {
                                xbaeDrawColumnLabel(mw, column, False);
                        }
                }
        }
}

static void xbaeRedrawFixed(XbaeMatrixWidget mw, Rectangle * expose, Boolean scrolled)
{
        Rectangle scrolled_expose, rect, intersect;
        int region;
        int row, start_row, end_row;
        int column, start_column, end_column;

        int cell_regions[] = {
                CLIP_FIXED_ROWS | CLIP_FIXED_COLUMNS,
                CLIP_FIXED_ROWS | CLIP_VISIBLE_WIDTH,
                CLIP_FIXED_ROWS | CLIP_TRAILING_FIXED_COLUMNS,
                CLIP_VISIBLE_HEIGHT | CLIP_FIXED_COLUMNS,
                CLIP_VISIBLE_HEIGHT | CLIP_TRAILING_FIXED_COLUMNS,
                CLIP_TRAILING_FIXED_ROWS | CLIP_FIXED_COLUMNS,
                CLIP_TRAILING_FIXED_ROWS | CLIP_VISIBLE_WIDTH,
                CLIP_TRAILING_FIXED_ROWS | CLIP_TRAILING_FIXED_COLUMNS
        };

        for (region = 0; region < sizeof cell_regions / sizeof *cell_regions; region++) {

                xbaeSmScrollRect(mw->matrix.matrix_scroll_mgr, expose, &scrolled_expose,
                                 cell_regions[region], scrolled);

                if (xbaeSetClipRect(mw, &rect, cell_regions[region])
                    && OVERLAP(scrolled_expose, rect)) {

                        SIMPLE_INTERSECT(scrolled_expose, rect, intersect);

                        DEBUGOUT(_XbaeDebug
                                 (__FILE__, (Widget) mw,
                                  "redrawing fixed cell region %d (x,y)=((%d,%d)-(%d,%d)) overlap (x,y)=((%d,%d)-(%d,%d))\n",
                                  region, rect.x1, rect.y1, rect.x2, rect.y2, intersect.x1,
                                  intersect.y1, intersect.x2, intersect.y2));

                        xbaeMatrixYtoRow(mw, &intersect.y1, &start_row);
                        xbaeMatrixYtoRow(mw, &intersect.y2, &end_row);

                        xbaeMatrixXtoColumn(mw, &intersect.x1, &start_column);
                        xbaeMatrixXtoColumn(mw, &intersect.x2, &end_column);

                        DEBUGOUT(_XbaeDebug
                                 (__FILE__, (Widget) mw,
                                  "redrawing fixed cell region %d (rows,columns)=((%d,%d)-(%d,%d))\n",
                                  region, start_row, start_column, end_row, end_column));

                        for (row = start_row; row <= end_row; row++) {
                                for (column = start_column; column <= end_column; column++) {
                                        xbaeDrawCell(mw, row, column);
                                }
                        }
                }
        }
}

/*
 * Redraw the row and column labels and the cells in fixed rows/columns
 * that are overlapped by the Rectangle argument.
 */
static void RedrawLabelsAndFixed(XbaeMatrixWidget mw, Rectangle * expose, Boolean scrolled)
{
        /*
         * Set up some local variables to avoid calling too many macros
         */
        Widget w = (Widget) mw;
        int horiz_sb_offset = HORIZ_SB_OFFSET(mw);
        int vert_sb_offset = VERT_SB_OFFSET(mw);
        int column_label_height = COLUMN_LABEL_HEIGHT(mw);
        int row_label_width = ROW_LABEL_WIDTH(mw);
        int trailing_fixed_row_height = TRAILING_FIXED_ROW_HEIGHT(mw);
        int fixed_column_width = FIXED_COLUMN_WIDTH(mw);
        int trailing_fixed_column_width = TRAILING_FIXED_COLUMN_WIDTH(mw);

        if (mw->matrix.disable_redisplay)
                return;

        DEBUGOUT(_XbaeDebug
                 (__FILE__, w,
                  "xbaeRedrawLabelsAndFixed() scroll_count=%d (offset_x,offset_y,count)=(%d,%d) with expose (x,y)=((%d,%d)-(%d,%d))\n",
                  mw->matrix.matrix_scroll_mgr->scroll_count,
                  mw->matrix.matrix_scroll_mgr->offset_x, mw->matrix.matrix_scroll_mgr->offset_y,
                  expose->x1, expose->y1, expose->x2, expose->y2));


        xbaeRedrawRowLabels(mw, expose, scrolled);
        xbaeRedrawColumnLabels(mw, expose, scrolled);
        xbaeRedrawFixed(mw, expose, scrolled);

        /* Danny Here */
        /*
         * Draw a shadow just inside row/column labels and around outer edge
         * of clip widget.  We can't use height of clip widget because it is
         * truncated to nearest row.  We use cell_visible_height instead.
         */
        if (mw->manager.shadow_thickness) {
                Dimension width, height;

                if (!mw->matrix.fill) {
                        width =
                            ClipChild(mw)->core.width + fixed_column_width +
                            trailing_fixed_column_width + 2 * mw->manager.shadow_thickness;
                        height =
                            mw->matrix.cell_visible_height + FIXED_ROW_HEIGHT(mw) +
                            trailing_fixed_row_height + 2 * mw->manager.shadow_thickness;
                } else {
                        width = mw->core.width - row_label_width - VERT_SB_SPACE(mw);

                        height = mw->core.height - column_label_height - HORIZ_SB_SPACE(mw);
                }

                DRAW_SHADOW(XtDisplay(mw), XtWindow(mw), "win", mw->manager.top_shadow_GC,
                            mw->manager.bottom_shadow_GC, mw->manager.shadow_thickness,
                            row_label_width + vert_sb_offset, column_label_height + horiz_sb_offset,
                            width, height, mw->matrix.shadow_type);
        }
}
void xbaeRedrawLabelsAndFixed(XbaeMatrixWidget mw, Rectangle * expose)
{
        RedrawLabelsAndFixed(mw, expose, False);
}

void xbaeRedrawScrolledLabelsAndFixed(XbaeMatrixWidget mw, Rectangle * expose)
{
        RedrawLabelsAndFixed(mw, expose, True);
}
